﻿Namespace NeuralRegression
	Friend Class NeuralRegressionProgram
		Shared Sub Main(ByVal args() As String)
			Console.WriteLine(vbLf & "Begin neural network regression demo" & vbLf)
			Console.WriteLine("Goal is to predict the sin(x)")

			' artificial; in realistic scenarios you'd read from a text file
			Dim numItems As Integer = 80
			Console.WriteLine(vbLf & "Programmatically generating " & numItems & " training data items")

			Dim trainData(numItems - 1)() As Double
			Dim rnd As New Random(1)

			For i As Integer = 0 To numItems - 1
				Dim x As Double = 6.4 * rnd.NextDouble() ' [0 to 2PI]
				Dim sx As Double = Math.Sin(x)
				trainData(i) = New Double() { x, sx }
			Next i

			Console.WriteLine(vbLf & "Training data:" & vbLf)
			ShowMatrix(trainData, 3, 4, True)

			Dim numInput As Integer = 1 ' usually more
			Dim numHidden As Integer = 12
			Dim numOutput As Integer = 1 ' usual for regression
			Dim rndSeed As Integer = 0

			Console.WriteLine(vbLf & "Creating a " & numInput & "-" & numHidden & "-" & numOutput & " regression neural network")
			Console.WriteLine("Using tanh hidden layer activation")
			Dim nn As New NeuralNetwork(numInput, numHidden, numOutput, rndSeed)

			Dim maxEpochs As Integer = 10000
			Dim learnRate As Double = 0.005
			Dim momentum As Double = 0.001
			Console.WriteLine(vbLf & "Setting maxEpochs = " & maxEpochs)
			Console.WriteLine("Setting learnRate = " & learnRate.ToString("F4"))
			Console.WriteLine("Setting momentum  = " & momentum.ToString("F4"))

			Console.WriteLine(vbLf & "Starting training (using stochastic back-propagation)")
			Dim weights() As Double = nn.Train(trainData, maxEpochs, learnRate, momentum)
			Console.WriteLine("Finished training")
			Console.WriteLine(vbLf & "Final neural network model weights:" & vbLf)
			ShowVector(weights, 4, 8, True)

			Dim y() As Double = nn.ComputeOutputs(New Double() { Math.PI })
			Console.WriteLine(vbLf & "Actual sin(PI)       =  0.0   Predicted =  " & y(0).ToString("F6"))

			y = nn.ComputeOutputs(New Double() { Math.PI / 2 })
			Console.WriteLine(vbLf & "Actual sin(PI / 2)   =  1.0   Predicted =  " & y(0).ToString("F6"))

			y = nn.ComputeOutputs(New Double() { 3 * Math.PI / 2.0 })
			Console.WriteLine(vbLf & "Actual sin(3*PI / 2) = -1.0   Predicted = " & y(0).ToString("F6"))

			y = nn.ComputeOutputs(New Double() { 6 * Math.PI })
			Console.WriteLine(vbLf & "Actual sin(6*PI)     =  0.0   Predicted =  " & y(0).ToString("F6"))

			Console.WriteLine(vbLf & "End demo" & vbLf)
			Console.ReadLine()
		End Sub ' Main

		Public Shared Sub ShowVector(ByVal vector() As Double, ByVal decimals As Integer, ByVal lineLen As Integer, ByVal newLine As Boolean)
			For i As Integer = 0 To vector.Length - 1
				If i > 0 AndAlso i Mod lineLen = 0 Then
					Console.WriteLine("")
				End If
				If vector(i) >= 0 Then
					Console.Write(" ")
				End If
				Console.Write(vector(i).ToString("F" & decimals) & " ")
			Next i
			If newLine = True Then
				Console.WriteLine("")
			End If
		End Sub

		Public Shared Sub ShowMatrix(ByVal matrix()() As Double, ByVal numRows As Integer, ByVal decimals As Integer, ByVal indices As Boolean)
			Dim len As Integer = matrix.Length.ToString().Length
			For i As Integer = 0 To numRows - 1
				If indices = True Then
					Console.Write("[" & i.ToString().PadLeft(len) & "]  ")
				End If
				For j As Integer = 0 To matrix(i).Length - 1
					Dim v As Double = matrix(i)(j)
					If v >= 0.0 Then
						Console.Write(" ") ' '+'
					End If
					Console.Write(v.ToString("F" & decimals) & "  ")
				Next j
				Console.WriteLine("")
			Next i

			If numRows < matrix.Length Then
				Console.WriteLine(". . .")
				Dim lastRow As Integer = matrix.Length - 1
				If indices = True Then
					Console.Write("[" & lastRow.ToString().PadLeft(len) & "]  ")
				End If
				For j As Integer = 0 To matrix(lastRow).Length - 1
					Dim v As Double = matrix(lastRow)(j)
					If v >= 0.0 Then
						Console.Write(" ") ' '+'
					End If
					Console.Write(v.ToString("F" & decimals) & "  ")
				Next j
			End If
			Console.WriteLine(vbLf)
		End Sub

	End Class ' Program

	Public Class NeuralNetwork
		Private numInput As Integer ' number input nodes
		Private numHidden As Integer
		Private numOutput As Integer

		Private inputs() As Double
		Private hiddens() As Double
		Private outputs() As Double

		Private ihWeights()() As Double ' input-hidden
		Private hBiases() As Double

		Private hoWeights()() As Double ' hidden-output
		Private oBiases() As Double

		Private rnd As Random

		Public Sub New(ByVal numInput As Integer, ByVal numHidden As Integer, ByVal numOutput As Integer, ByVal seed As Integer)
			Me.numInput = numInput
			Me.numHidden = numHidden
			Me.numOutput = numOutput

			Me.inputs = New Double(numInput - 1){}
			Me.hiddens = New Double(numHidden - 1){}
			Me.outputs = New Double(numOutput - 1){}


			Me.ihWeights = MakeMatrix(numInput, numHidden, 0.0)
			Me.hBiases = New Double(numHidden - 1){}

			Me.hoWeights = MakeMatrix(numHidden, numOutput, 0.0)
			Me.oBiases = New Double(numOutput - 1){}

			Me.rnd = New Random(seed)
			Me.InitializeWeights() ' all weights and biases
		End Sub ' ctor

		Private Shared Function MakeMatrix(ByVal rows As Integer, ByVal cols As Integer, ByVal v As Double) As Double()() ' helper for ctor, Train
			Dim result(rows - 1)() As Double
			For r As Integer = 0 To result.Length - 1
				result(r) = New Double(cols - 1){}
			Next r
			For i As Integer = 0 To rows - 1
				For j As Integer = 0 To cols - 1
					result(i)(j) = v
				Next j
			Next i
			Return result
		End Function

		Private Sub InitializeWeights() ' helper for ctor
			' initialize weights and biases to random values between 0.0001 and 0.001
			Dim numWeights As Integer = (numInput * numHidden) + (numHidden * numOutput) + numHidden + numOutput
			Dim initialWeights(numWeights - 1) As Double
			Dim lo As Double = -0.001
			Dim hi As Double = +0.001
			For i As Integer = 0 To initialWeights.Length - 1
				initialWeights(i) = (hi - lo) * rnd.NextDouble() + lo ' [-0.001 to +0.001]
			Next i
																		'initialWeights[i] = (0.001 - 0.0001) * rnd.NextDouble() + 0.0001;
			Me.SetWeights(initialWeights)
		End Sub

		Public Sub SetWeights(ByVal weights() As Double)
			' copy serialized weights and biases in weights[] array
			' to i-h weights, i-h biases, h-o weights, h-o biases
			Dim numWeights As Integer = (numInput * numHidden) + (numHidden * numOutput) + numHidden + numOutput
			If weights.Length <> numWeights Then
				Throw New Exception("Bad weights array in SetWeights")
			End If

			Dim w As Integer = 0 ' points into weights param

			For i As Integer = 0 To numInput - 1
				For j As Integer = 0 To numHidden - 1
					ihWeights(i)(j) = weights(w)
					w += 1
				Next j
			Next i

			For j As Integer = 0 To numHidden - 1
				hBiases(j) = weights(w)
				w += 1
			Next j

			For j As Integer = 0 To numHidden - 1
				For k As Integer = 0 To numOutput - 1
					hoWeights(j)(k) = weights(w)
					w += 1
				Next k
			Next j

			For k As Integer = 0 To numOutput - 1
				oBiases(k) = weights(k)
				k += 1
			Next k
		End Sub

		Public Function GetWeights() As Double()
			Dim numWeights As Integer = (numInput * numHidden) + (numHidden * numOutput) + numHidden + numOutput
			Dim result(numWeights - 1) As Double

			Dim w As Integer = 0
			For i As Integer = 0 To numInput - 1
				For j As Integer = 0 To numHidden - 1
					result(w) = ihWeights(i)(j)
					w += 1
				Next j
			Next i

			For j As Integer = 0 To numHidden - 1
				result(w) = hBiases(j)
				w += 1
			Next j

			For j As Integer = 0 To numHidden - 1
				For k As Integer = 0 To numOutput - 1
					result(w) = hoWeights(j)(k)
					w += 1
				Next k
			Next j

			For k As Integer = 0 To numOutput - 1
				result(w) = oBiases(k)
				w += 1
			Next k

			Return result
		End Function

		Public Function ComputeOutputs(ByVal xValues() As Double) As Double()
			Dim hSums(numHidden - 1) As Double ' hidden nodes sums scratch array
			Dim oSums(numOutput - 1) As Double ' output nodes sums

			For i As Integer = 0 To numInput - 1 ' copy x-values to inputs
				Me.inputs(i) = xValues(i)
			Next i

			For j As Integer = 0 To numHidden - 1 ' compute i-h sum of weights * inputs
				For i As Integer = 0 To numInput - 1
					hSums(j) += Me.inputs(i) * Me.ihWeights(i)(j) ' note +=
				Next i
			Next j

			For j As Integer = 0 To numHidden - 1 ' add biases to input-to-hidden sums
				hSums(j) += Me.hBiases(j)
			Next j

			For j As Integer = 0 To numHidden - 1 ' apply activation
				Me.hiddens(j) = HyperTan(hSums(j)) ' hard-coded
			Next j

			For k As Integer = 0 To numOutput - 1 ' compute h-o sum of weights * hOutputs
				For j As Integer = 0 To numHidden - 1
					oSums(k) += hiddens(j) * hoWeights(j)(k)
				Next j
			Next k

			For k As Integer = 0 To numOutput - 1 ' add biases to input-to-hidden sums
				oSums(k) += oBiases(k)
			Next k

			Array.Copy(oSums, Me.outputs, outputs.Length) ' copy without activation

			Dim retResult(numOutput - 1) As Double ' could define a GetOutputs
			Array.Copy(Me.outputs, retResult, retResult.Length)
			Return retResult
		End Function

		Private Shared Function HyperTan(ByVal x As Double) As Double
			If x < -20.0 Then ' approximation is correct to 30 decimals
				Return -1.0
			ElseIf x > 20.0 Then
				Return 1.0
			Else
				Return Math.Tanh(x)
			End If
		End Function

		Public Function Train(ByVal trainData()() As Double, ByVal maxEpochs As Integer, ByVal learnRate As Double, ByVal momentum As Double) As Double()
			' train using back-prop
			' back-prop specific arrays
			Dim hoGrads()() As Double = MakeMatrix(numHidden, numOutput, 0.0) ' hidden-to-output weights gradients
			Dim obGrads(numOutput - 1) As Double ' output biases gradients

			Dim ihGrads()() As Double = MakeMatrix(numInput, numHidden, 0.0) ' input-to-hidden weights gradients
			Dim hbGrads(numHidden - 1) As Double ' hidden biases gradients

			Dim oSignals(numOutput - 1) As Double ' signals == gradients w/o associated input terms
			Dim hSignals(numHidden - 1) As Double ' hidden node signals

			' back-prop momentum specific arrays 
			Dim ihPrevWeightsDelta()() As Double = MakeMatrix(numInput, numHidden, 0.0)
			Dim hPrevBiasesDelta(numHidden - 1) As Double
			Dim hoPrevWeightsDelta()() As Double = MakeMatrix(numHidden, numOutput, 0.0)
			Dim oPrevBiasesDelta(numOutput - 1) As Double

			' train a back-prop style NN regression using learning rate and momentum
			Dim epoch As Integer = 0
			Dim xValues(numInput - 1) As Double ' inputs
			Dim tValues(numOutput - 1) As Double ' target values

			Dim sequence(trainData.Length - 1) As Integer
			For i As Integer = 0 To sequence.Length - 1
				sequence(i) = i
			Next i

			Dim errInterval As Integer = maxEpochs \ 10 ' interval to check validation data
			Do While epoch < maxEpochs
				epoch += 1 ' immediately to prevent display when 0

				If epoch Mod errInterval = 0 AndAlso epoch < maxEpochs Then
					Dim trainErr As Double = [Error](trainData)
					Console.WriteLine("epoch = " & epoch & "  training error = " & trainErr.ToString("F4"))
				End If

				Shuffle(sequence) ' visit each training data in random order
				For ii As Integer = 0 To trainData.Length - 1
					Dim idx As Integer = sequence(ii)
					Array.Copy(trainData(idx), xValues, numInput)
					Array.Copy(trainData(idx), numInput, tValues, 0, numOutput)
					ComputeOutputs(xValues) ' copy xValues in, compute outputs

					' indices: i = inputs, j = hiddens, k = outputs

					' 1. compute output nodes signals (assumes constant activation)
					For k As Integer = 0 To numOutput - 1
						Dim derivative As Double = 1.0 ' for dummy output activation f'
						oSignals(k) = (tValues(k) - outputs(k)) * derivative
					Next k

					' 2. compute hidden-to-output weights gradients using output signals
					For j As Integer = 0 To numHidden - 1
						For k As Integer = 0 To numOutput - 1
							hoGrads(j)(k) = oSignals(k) * hiddens(j)
						Next k
					Next j

					' 2b. compute output biases gradients using output signals
					For k As Integer = 0 To numOutput - 1
						obGrads(k) = oSignals(k) * 1.0 ' dummy assoc. input value
					Next k

					' 3. compute hidden nodes signals
					For j As Integer = 0 To numHidden - 1
						Dim sum As Double = 0.0 ' need sums of output signals times hidden-to-output weights
						For k As Integer = 0 To numOutput - 1
							sum += oSignals(k) * hoWeights(j)(k)
						Next k
						Dim derivative As Double = (1 + hiddens(j)) * (1 - hiddens(j)) ' for tanh
						hSignals(j) = sum * derivative
					Next j

					' 4. compute input-hidden weights gradients
					For i As Integer = 0 To numInput - 1
						For j As Integer = 0 To numHidden - 1
							ihGrads(i)(j) = hSignals(j) * inputs(i)
						Next j
					Next i

					' 4b. compute hidden node biases gradienys
					For j As Integer = 0 To numHidden - 1
						hbGrads(j) = hSignals(j) * 1.0 ' dummy 1.0 input
					Next j

					' == update weights and biases

					' 1. update input-to-hidden weights
					For i As Integer = 0 To numInput - 1
						For j As Integer = 0 To numHidden - 1
							Dim delta As Double = ihGrads(i)(j) * learnRate
							ihWeights(i)(j) += delta
							ihWeights(i)(j) += ihPrevWeightsDelta(i)(j) * momentum
							ihPrevWeightsDelta(i)(j) = delta ' save for next time
						Next j
					Next i

					' 2. update hidden biases
					For j As Integer = 0 To numHidden - 1
						Dim delta As Double = hbGrads(j) * learnRate
						hBiases(j) += delta
						hBiases(j) += hPrevBiasesDelta(j) * momentum
						hPrevBiasesDelta(j) = delta
					Next j

					' 3. update hidden-to-output weights
					For j As Integer = 0 To numHidden - 1
						For k As Integer = 0 To numOutput - 1
							Dim delta As Double = hoGrads(j)(k) * learnRate
							hoWeights(j)(k) += delta
							hoWeights(j)(k) += hoPrevWeightsDelta(j)(k) * momentum
							hoPrevWeightsDelta(j)(k) = delta
						Next k
					Next j

					' 4. update output node biases
					For k As Integer = 0 To numOutput - 1
						Dim delta As Double = obGrads(k) * learnRate
						oBiases(k) += delta
						oBiases(k) += oPrevBiasesDelta(k) * momentum
						oPrevBiasesDelta(k) = delta
					Next k

				Next ii ' each training item

			Loop ' while
			Dim bestWts() As Double = Me.GetWeights()
			Return bestWts
		End Function ' Train


		Private Sub Shuffle(ByVal sequence() As Integer) ' an instance method
			For i As Integer = 0 To sequence.Length - 1
				Dim r As Integer = Me.rnd.Next(i, sequence.Length)
				Dim tmp As Integer = sequence(r)
				sequence(r) = sequence(i)
				sequence(i) = tmp
			Next i
		End Sub ' Shuffle


		Private Function [Error](ByVal data()() As Double) As Double
			' MSE == average squared error per training item
			Dim sumSquaredError As Double = 0.0
			Dim xValues(numInput - 1) As Double ' first numInput values in trainData
			Dim tValues(numOutput - 1) As Double ' last numOutput values

			' walk thru each training case
			For i As Integer = 0 To data.Length - 1
				Array.Copy(data(i), xValues, numInput)
				Array.Copy(data(i), numInput, tValues, 0, numOutput) ' get target value(s)
				Dim yValues() As Double = Me.ComputeOutputs(xValues) ' outputs using current weights

				For j As Integer = 0 To numOutput - 1
					Dim err As Double = tValues(j) - yValues(j)
					sumSquaredError += err * err
				Next j
			Next i
			Return sumSquaredError / data.Length
		End Function ' Error

	End Class ' class NeuralNetwork

End Namespace ' ns